# Copyright (C) 2021 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
Test module for galene_stream.webrtc.
"""

import asyncio

from galene_stream.webrtc import WebRTCClient


def test_init_webrtc():
    """Test WebRTC initialization."""
    event_loop = asyncio.get_event_loop()
    client = WebRTCClient("rtmp://localhost:1935/live/test", 1048576, None, None)
    client.start_pipeline(event_loop, [])
    client.close_pipeline()



# Copyright (C) 2021 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
Main command-line script for Galène stream gateway.
"""

import argparse
import asyncio
import logging
import sys

from galene_stream.galene import GaleneClient


def start(opt: argparse.Namespace):
    """Init Galène client and start gateway

    :param opt: program options
    :type opt: argparse.Namespace
    """
    client = GaleneClient(
        opt.input, opt.output, opt.bitrate, opt.group, opt.username, opt.password
    )

    # Connect and run main even loop
    event_loop = asyncio.get_event_loop()
    event_loop.run_until_complete(client.connect())
    try:
        event_loop.run_until_complete(client.loop(event_loop))
        event_loop.run_until_complete(client.close())
    except KeyboardInterrupt:
        event_loop.run_until_complete(client.close())
        sys.exit(1)


def main():
    """Entrypoint."""
    # Arguments parser
    parser = argparse.ArgumentParser(
        prog="galene-stream",
        description="Galène stream gateway.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="debug mode: show debug messages",
    )
    parser.add_argument(
        "-i",
        "--input",
        required=True,
        help=(
            'URI to use as GStreamer "uridecodebin" module input, '
            'e.g. "rtmp://localhost:1935/live/test"'
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        required=True,
        help='Galène server to connect to, e.g. "wss://galene.example.com/ws"',
    )
    parser.add_argument(
        "-b",
        "--bitrate",
        default=1048576,
        help="VP8 encoder bitrate in bit/s, you should adapt this to your network, default to 1048576",
    )
    parser.add_argument(
        "-g",
        "--group",
        required=True,
        help="Join this group",
    )
    parser.add_argument(
        "-u",
        "--username",
        required=True,
        help="Group username",
    )
    parser.add_argument(
        "-p",
        "--password",
        help="Group password",
    )
    options = parser.parse_args()

    # Configure logging
    level = logging.DEBUG if options.debug else logging.INFO
    logging.addLevelName(logging.INFO, "\033[1;36mINFO\033[1;0m")
    logging.addLevelName(logging.WARNING, "\033[1;33mWARNING\033[1;0m")
    logging.addLevelName(logging.ERROR, "\033[1;91mERROR\033[1;0m")
    logging.addLevelName(logging.DEBUG, "\033[1;30mDEBUG")
    logging.basicConfig(
        level=level,
        format="\033[90m%(asctime)s\033[1;0m [%(name)s] %(levelname)s %(message)s\033[1;0m",
    )

    start(options)



# Copyright (C) 2021 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
WebRTC support using Gstreamer.
"""

import asyncio
import logging
import os
import pprint
import sys
from typing import List

import gi

# Specify version, then import gstreamer plugins
gi.require_version("Gst", "1.0")
gi.require_version("GstWebRTC", "1.0")
gi.require_version("GstSdp", "1.0")

from gi.repository import Gst, GstSdp, GstWebRTC

log = logging.getLogger(__name__)


class WebRTCClient:
    """WebRTCClient

    Based on <https://gitlab.freedesktop.org/gstreamer/gst-examples/>.
    """

    def __init__(
        self, input_uri: str, bitrate: int, sdp_offer_callback, ice_candidate_callback
    ) -> None:
        """Init WebRTCClient.

        :param input_uri: URI for GStreamer uridecodebin
        :type input_uri: str
        :param bitrate: VP8 encoder bitrate in bit/s
        :type bitrate: int
        :param sdp_offer_callback: coroutine to send SDP offer
        :type sdp_offer_callback: coroutine
        :param ice_candidate_callback: coroutine to send ICE candidate
        :type ice_candidate_callback: coroutine
        """
        self.event_loop = None
        self.pipe = None
        self.webrtc = None
        self.sdp_offer_callback = sdp_offer_callback
        self.ice_candidate_callback = ice_candidate_callback

        self.pipeline_desc = (
            "webrtcbin name=send bundle-policy=max-bundle "
            f'uridecodebin uri="{input_uri}" name=bin '
            f"bin. ! videoconvert ! vp8enc deadline=1 target-bitrate={bitrate} ! rtpvp8pay pt=97 ! send. "
            "bin. ! audioconvert ! audioresample ! opusenc ! rtpopuspay pt=96 ! send."
        )

        # If gstreamer debug level is undefined, show warnings and errors
        if "GST_DEBUG" not in os.environ:
            os.environ["GST_DEBUG"] = "2"

        # Initialize GStreamer and check available plugins
        Gst.init(None)
        needed = [
            "opus",
            "vpx",
            "nice",
            "webrtc",
            "dtls",
            "srtp",
            "rtp",
            "rtpmanager",
            "x264",
        ]
        missing = filter(lambda p: Gst.Registry.get().find_plugin(p) is None, needed)
        missing_list = list(missing)
        if len(missing_list):
            log.error(f"Missing gstreamer plugins: {missing_list}")
            sys.exit(1)

    def on_offer_created(self, promise, _, __) -> None:
        """``on-offer-created`` event handler.

        :param promise: promise running this event
        :type promise: Gst.Promise
        """
        assert self.event_loop is not None
        assert self.webrtc is not None

        # Get offer from the promise calling the event
        promise.wait()
        reply = promise.get_reply()
        offer = reply.get_value("offer")

        # Set local description
        log.info("Setting local description")
        promise = Gst.Promise.new()
        self.webrtc.emit("set-local-description", offer, promise)
        promise.interrupt()

        # Send local SDP offer to remote
        offer = offer.sdp.as_text()
        future = asyncio.run_coroutine_threadsafe(
            self.sdp_offer_callback(offer), self.event_loop
        )
        future.result()  # wait

    def on_negotiation_needed(self, element) -> None:
        """``on-negotiation-needed`` event handler.

        When receiving ``on-negotiation-needed`` event, create new offer.

        :param element: the webrtcbin
        :type element: object
        """
        # Set up ``on-offer-created`` event when offer is ready
        promise = Gst.Promise.new_with_change_func(self.on_offer_created, element, None)

        # Create new offer
        element.emit("create-offer", None, promise)

    def on_ice_candidate(self, _, mline_index, candidate: str) -> None:
        """``on-ice-candidate`` event handler.

        Send ICE candidate message to remote.

        :param mline_index: the index of the media description in the SDP
        :type mline_index: str
        :param candidate: an ICE candidate
        :type candidate: str
        """
        assert self.event_loop is not None

        c = {"candidate": candidate, "sdpMLineIndex": mline_index}
        future = asyncio.run_coroutine_threadsafe(
            self.ice_candidate_callback(c), self.event_loop
        )
        future.result()  # wait

    def set_remote_sdp(self, sdp: str) -> None:
        """Set remote session description.

        :param sdp: Session description
        :type sdp: str
        """
        assert self.webrtc is not None

        log.info("Setting remote session description")
        _, sdp_msg = GstSdp.SDPMessage.new()
        GstSdp.sdp_message_parse_buffer(bytes(sdp.encode()), sdp_msg)
        answer = GstWebRTC.WebRTCSessionDescription.new(
            GstWebRTC.WebRTCSDPType.ANSWER, sdp_msg
        )
        promise = Gst.Promise.new()
        self.webrtc.emit("set-remote-description", answer, promise)
        promise.interrupt()

    def add_ice_candidate(self, mline_index: int, candidate: str) -> None:
        """Add new ICE candidate.

        :param mline_index: the index of the media description in the SDP
        :type mline_index: int
        :param candidate: an ice candidate
        :type candidate: str
        """
        assert self.webrtc is not None
        self.webrtc.emit("add-ice-candidate", mline_index, candidate)

    def get_stats(self) -> str:
        """Get RTP statistics from GStreamer.

        :return: statistics as text report
        :rtype: str
        """
        assert self.pipe is not None
        fields = [
            "ssrc",
            "is-sender",
            "clock-rate",
            "octets-sent",
            "packets-sent",
            "octets-received",
            "packets-received",
            "bitrate",
            "packets-lost",
            "recv-pli-count",
            "recv-nack-count",
            "sr-ntptime",
        ]
        rtpbin = self.pipe.get_by_name("rtpsession0")
        message = []
        if rtpbin is None:
            return ""

        # Get statistics for each SSRC
        stats = rtpbin.get_property("stats")
        sources_stats = stats.get_value("source-stats")
        for source_stats in sources_stats:
            if source_stats.get_value("ssrc") != 0:
                message.append({f: source_stats.get_value(f) for f in fields})
        return pprint.pformat(message, sort_dicts=False)

    def start_pipeline(
        self, event_loop: asyncio.AbstractEventLoop, ice_servers: List[str]
    ) -> None:
        """Start gstreamer pipeline and connect WebRTC events.

        :param event_loop: asyncio event loop
        :type event_loop: EventLoop
        :param ice_servers: list of ICE TURN servers
        :type ice_servers: list of str
        """
        log.info("Starting pipeline")
        self.event_loop = event_loop
        self.pipe = Gst.parse_launch(self.pipeline_desc)
        self.webrtc = self.pipe.get_by_name("send")
        self.webrtc.connect("on-negotiation-needed", self.on_negotiation_needed)
        self.webrtc.connect("on-ice-candidate", self.on_ice_candidate)

        # Enable WebRTC negative acknowledgement and FEC
        transceiver_count = self.webrtc.emit("get-transceivers").len
        for i in range(transceiver_count):
            transceiver = self.webrtc.emit("get-transceiver", i)
            transceiver.set_property("do-nack", True)
            # ULPFEC is not yet supported by Galène
            # transceiver.set_property("fec-type", GstWebRTC.WebRTCFECType.ULP_RED)

        # Add TURN servers
        try:
            for uri in ice_servers:
                self.webrtc.emit("add-turn-server", uri)
        except TypeError:
            log.warn(
                "add-turn-server signal is missing, maybe your gstreamer "
                "is too old. Skipping TURN servers configuration"
            )

        # Start
        self.pipe.set_state(Gst.State.PLAYING)

    def close_pipeline(self) -> None:
        """Stop gstreamer pipeline."""
        log.info("Closing pipeline")

        # If pipeline is running, then export pipeline graph before closing
        # To use this, set GST_DEBUG_DUMP_DOT_DIR environnement variable
        if self.pipe is not None:
            Gst.debug_bin_to_dot_file_with_ts(
                self.pipe, Gst.DebugGraphDetails.ALL, "pipeline"
            )
            self.pipe.set_state(Gst.State.NULL)

        self.pipe = None
        self.webrtc = None



# Copyright (C) 2021 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
Galène protocol support.
"""

import json
import logging
import secrets
from typing import List

import websockets

from galene_stream.webrtc import WebRTCClient

log = logging.getLogger(__name__)


class GaleneClient:
    """Galène protocol implementation."""

    def __init__(
        self,
        input_uri: str,
        server: str,
        bitrate: int,
        group: str,
        username: str,
        password: str = "",
    ) -> None:
        """Create GaleneClient

        :param input_uri: URI for GStreamer uridecodebin
        :type input_uri: str
        :param server: websocket url to connect to
        :type server: str
        :param bitrate: VP8 encoder bitrate in bit/s
        :type bitrate: int
        :param group: group to join
        :type group: str
        :param username: group user name
        :type username: str
        :param password: group user password if required
        :type password: str, optional
        """
        self.server = server
        self.group = group
        self.username = username
        self.password = password

        self.conn = None
        self.ice_servers: List[str] = []
        self.client_id = secrets.token_bytes(16).hex()
        self.webrtc = WebRTCClient(
            input_uri, bitrate, self.send_sdp_offer, self.send_ice_candidate
        )

    async def send(self, message: dict) -> None:
        """Send message to remote.

        :param message: message to send
        :type message: dict
        """
        msg = json.dumps(message)
        if self.conn is None:
            log.error("Connection is closed, cannot send message")
            return
        await self.conn.send(msg)

    async def send_sdp_offer(self, sdp: str) -> None:
        """Send SDP offer to remote.

        :param sdp: session description
        :type sdp: str
        """
        log.debug(f"Sending local SDP offer to remote: {sdp}")
        # We use client_id as stream id, but it can differ
        msg = {
            "type": "offer",
            "id": self.client_id,  # stream id
            "source": self.client_id,
            "username": self.username,
            "sdp": sdp,
            "label": "video",
        }
        await self.send(msg)

    async def send_ice_candidate(self, candidate: dict) -> None:
        """Send ICE candidate to remote.

        :param canditate: ICE candidate
        :type canditate: dict
        """
        log.debug("Sending new ICE candidate to remote")
        msg = {"type": "ice", "id": self.client_id, "candidate": candidate}
        await self.send(msg)

    async def send_chat(self, message: str) -> None:
        """Send chat message.

        :param message: content of the message
        :type message: str
        """
        await self.send(
            {
                "type": "chat",
                "source": self.client_id,
                "username": self.username,
                "noecho": True,
                "value": message,
            }
        )

    async def connect(self) -> None:
        """Connect to server."""
        # Create WebSocket
        log.info("Connecting to WebSocket")
        self.conn = await websockets.connect(self.server, ssl=True)

        # Handshake with server
        log.info("Handshaking")
        msg = {
            "type": "handshake",
            "version": ["1"],  # since Galene 0.6.0
            "id": self.client_id,
        }
        await self.send(msg)
        await self.conn.recv()  # wait for handshake

        # Join group
        log.info("Joining group")
        msg = {
            "type": "join",
            "kind": "join",
            "group": self.group,
            "username": self.username,
            "password": self.password,
        }
        await self.send(msg)
        response = {"type": "none"}
        while response["type"] != "joined":
            # The server will send 'user' messages that we ignore
            raw_response = await self.conn.recv()
            response = json.loads(raw_response)
        if response["kind"] != "join":
            raise RuntimeError("failed to join room")

        # Get ICE servers
        rtc_configuration = response.get("rtcConfiguration", {})
        assert isinstance(rtc_configuration, dict)
        self.ice_servers = []
        for server in rtc_configuration.get("iceServers", []):
            username = server.get("username", "")
            credential = server.get("credential", "")
            for url in server.get("urls", []):
                url = url.replace("turn:", "")  # remove prefix
                uri = f"turn://{username}:{credential}@{url}"
                self.ice_servers.append(uri)

    async def close(self) -> None:
        """Close connection."""
        log.info("Closing WebSocket connection")
        self.webrtc.close_pipeline()
        if self.conn is None:
            log.warn("Connection is already closed")
            return
        await self.conn.close()
        self.conn = None

    async def loop(self, event_loop) -> None:
        """Client loop

        :param event_loop: asyncio event loop
        :type event_loop: EventLoop
        :raises RuntimeError: if client is not connected
        """
        if self.conn is None:
            raise RuntimeError("client not connected")
        self.webrtc.start_pipeline(event_loop, self.ice_servers)
        log.info("Waiting for incoming stream...")

        async for message in self.conn:
            message = json.loads(message)
            if message["type"] == "ping":
                # Need to answer pong to ping request to keep connection
                await self.send({"type": "pong"})
            elif message["type"] == "abort":
                # Server wants to close our stream
                log.info("Received abort from server")
                await self.send({"type": "close", "id": message.get("id")})
                break
            elif message["type"] == "answer":
                # Server is sending a SDP offer
                sdp = message.get("sdp")
                log.debug(f"Receiving SDP from remote: {sdp}")
                self.webrtc.set_remote_sdp(sdp)
            elif message["type"] == "ice":
                # Server is sending trickle ICE candidates
                log.debug("Receiving new ICE candidate from remote")
                mline_index = message.get("candidate").get("sdpMLineIndex")
                candidate = message.get("candidate").get("candidate")
                self.webrtc.add_ice_candidate(mline_index, candidate)
            elif message["type"] == "renegotiate":
                # Server is asking to renegotiate WebRTC session
                self.webrtc.on_negotiation_needed(self.webrtc.webrtc)
            elif message["type"] == "usermessage":
                value = message.get("value")
                if message["kind"] == "error":
                    log.error(f"Server returned error: {value}")
                    break
                else:
                    log.warn(f"Server sent: {value}")
            elif message["type"] == "user":
                continue  # ignore user events
            elif message["type"] == "close":
                continue  # ignore close events
            elif message["type"] == "chat":
                # User might request statistics using `!webrtc` chat command
                if message.get("value") == "!webrtc":
                    m = self.webrtc.get_stats()
                    if m:
                        await self.send_chat(m)
            elif message["type"] == "chathistory":
                continue  # ignore chat history events
            else:
                # Oh no! We receive something not implemented
                log.warn(f"Not implemented {message}")



#!/usr/bin/env python3
# Copyright (C) 2022 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
Entrypoint for Python module
"""

from galene_stream.cli import main

if __name__ == "__main__":
    main()



# Copyright (C) 2021 Alexandre Iooss
# SPDX-License-Identifier: MIT

"""
Galène stream gateway.
"""

try:
    from galene_stream.version import version as __version__
except ImportError:
    __version__ = "dev"

# See https://www.python.org/dev/peps/pep-0008/#module-level-dunder-names
__all__ = [
    "__version__",
]
